"""Internal function to extract timeseries with or without multiprocessing"""
import copy, json, math, os, re, warnings
import numpy as np, pandas as pd
from nilearn.maskers import NiftiLabelsMasker
from nilearn.image import index_img, load_img

def _extract_timeseries(subj_id, nifti_files, mask_files, event_files, confound_files, confound_metadata_files,
                        run_list, tr, condition, parcel_approach, signal_clean_info, verbose, flush_print, task_info):

    # Initialize subject dictionary
    subject_timeseries = {subj_id: {}}

    for run in run_list:
        # Initialize flag; This is for flagging runs where the percentage of volumes exceeding the "fd_threshold"
        # is greater than "outlier_percentage"
        flagged = False

        # Due to prior checking in _setup_extraction within TimeseriesExtractor, run_list = [None] is assumed to be a
        # single file without the run- description
        run_id = "run-0" if run is None else run
        run = run if run is not None else ""

        # Get files from specific run
        nifti_file = [nifti_file for nifti_file in nifti_files if run in os.path.basename(nifti_file)][0]
        if mask_files:
            mask_file = [mask_file for mask_file in mask_files if run in os.path.basename(mask_file)][0]
        else:
            mask_file = None
        confound_file = [confound_file for confound_file in confound_files
                         if run in os.path.basename(confound_file)][0] if signal_clean_info["use_confounds"] else None

        if signal_clean_info["use_confounds"] and signal_clean_info["n_acompcor_separate"]:
            confound_metadata_file = [confound_metadata_file for confound_metadata_file
                                      in confound_metadata_files if run in os.path.basename(confound_metadata_file)][0]
        else: confound_metadata_file = None

        confound_df = pd.read_csv(confound_file, sep="\t") if signal_clean_info["use_confounds"] else None

        event_file = None if not event_files else [event_file for event_file in event_files
                                                   if run in os.path.basename(event_file)][0]

        # Base message
        run_message = run_id.split("-")[-1]
        if task_info["session"] is not None:
            session_message = task_info["session"]
        else:
            base_filename = os.path.basename(nifti_file)
            session_id = re.search("ses-(\\S+?)[-_]",base_filename)[0][:-1] if "ses-" in base_filename else None
            session_message = session_id.split("-")[-1] if session_id else None
        sub_message = f'[SUBJECT: {subj_id} | SESSION: {session_message} | TASK: {task_info["task"]} | RUN: {run_message}]'
        underline = '-'*len(sub_message)
        subject_header = f"\n{sub_message}\n{underline}\n"

        if verbose:
            print(f"{subject_header}" + f"Preparing for timeseries extraction using - [FILE: {nifti_file}].",
                  flush=flush_print)

        # Initialize variables; fd_threshold and even_file checked first for a "fail fast" approach that avoids
        # extracting timeseries for runs that will be empty or flagged
        scan_list, censor_volumes = [], []
        censor = False
        threshold, outlier_limit = None, None
        dummy_scans = None

        # Check for non-steady_state if requested
        if signal_clean_info["dummy_scans"]:
            if isinstance(signal_clean_info["dummy_scans"], dict) and signal_clean_info["dummy_scans"]["auto"] is True:
                dummy_scans, dummy_flag = len([col for col in confound_df.columns if "non_steady_state" in col]), "auto"
                if "min" in signal_clean_info["dummy_scans"] and dummy_scans < signal_clean_info["dummy_scans"]["min"]:
                    dummy_scans, dummy_flag = signal_clean_info["dummy_scans"]["min"], "min"
                if "max" in signal_clean_info["dummy_scans"] and dummy_scans > signal_clean_info["dummy_scans"]["max"]:
                    dummy_scans, dummy_flag = signal_clean_info["dummy_scans"]["max"], "max"
                if dummy_scans == 0: dummy_scans = None
                if verbose:
                    if dummy_flag == "auto":
                        if dummy_scans is not None:
                            print(f"{subject_header}" + "Number of dummy scans to be removed based on "
                                  f"'non_steady_state_outlier_XX' columns is {dummy_scans}.", flush=flush_print)
                        else:
                            print(f"{subject_header}" + "No 'non_steady_state_outlier_XX' columns were found so 0 "
                                  "dummy scans will be removed.", flush=flush_print)
                    else:
                        print(f"{subject_header}" + f"Default dummy scans set by {dummy_flag} will be used: {dummy_scans}",
                              flush=flush_print)
            else:
                dummy_scans = signal_clean_info["dummy_scans"]

        if signal_clean_info["use_confounds"] and signal_clean_info["fd_threshold"]:
            if "framewise_displacement" in confound_df.columns:
                censor = True
                fd_array = confound_df["framewise_displacement"].fillna(0).values
                # Truncate fd_array if dummy scans; done to only assess if the truncated array meets the "outlier_percentage" criteria
                if dummy_scans: fd_array = fd_array[dummy_scans:]

                # Check if float or dict
                if isinstance(signal_clean_info["fd_threshold"], dict):
                    threshold = signal_clean_info["fd_threshold"]["threshold"]
                    if "outlier_percentage" in signal_clean_info["fd_threshold"]:
                        outlier_limit = signal_clean_info["fd_threshold"]["outlier_percentage"]
                else:
                    threshold = signal_clean_info["fd_threshold"]

                censor_volumes.extend(list(np.where(fd_array > threshold)[0]))

                if len(censor_volumes) == fd_array.shape[0]:
                    warnings.warn(f"{subject_header}" + "Processing skipped: Timeseries will be empty due to the "
                                  f"framewise displacement of all volumes exceeding {threshold}.")
                    continue

                if outlier_limit and condition is None:
                    # Determine if run is flagged; when condition is not specified
                    percentage = len(censor_volumes)/len(fd_array)
                    flagged = True if percentage > outlier_limit else False
            else:
                warnings.warn(f"{subject_header}" + "`fd_threshold` specified but 'framewise_displacement' column not "
                              "in the confound tsv file. Removal of volumes after nuisance regression will not be done.")

        if event_file:
            event_df = pd.read_csv(event_file, sep="\t")
            # Get specific timing information for specific condition
            condition_df = event_df[event_df["trial_type"] == condition]

            if condition_df.empty:
                warnings.warn(f"{subject_header}" + f"[CONDITION: {condition}] - Processing skipped: Condition does not "
                              "exist in the 'trial_type' column of the event file.")
                continue

            # Convert times into scan numbers to obtain the scans taken when the participant was exposed to the
            # condition of interest; include partial scans
            for i in condition_df.index:
                onset_scan = int(condition_df.loc[i,"onset"]/tr)
                end_scan = math.ceil((condition_df.loc[i,"onset"] + condition_df.loc[i,"duration"])/tr)
                # Add one since range is not inclusive
                scan_list.extend(list(range(onset_scan, end_scan + 1)))

            # Get unique scans to not duplicate information
            scan_list = sorted(list(set(scan_list)))

            # Adjust for dummy before censoring since censoring is dummy adjusted above
            if dummy_scans: scan_list = [scan - dummy_scans for scan in scan_list if scan not in range(0, dummy_scans)]

            if censor:
                # Assess if any condition indices greater than fd array to not dilute outlier calculation
                if max(scan_list) > fd_array.shape[0] - 1:
                    scan_list = _check_indices(scan_list,fd_array.shape[0],condition,subject_header)
                # Get length of scan list prior to assess outliers if requested
                before_censor = len(scan_list)
                scan_list = [volume for volume in scan_list if volume not in censor_volumes]
                if outlier_limit:
                    percentage = 1 - (len(scan_list)/before_censor)
                    flagged = True if percentage > outlier_limit else False

            if not scan_list:
                warnings.warn(f"{subject_header}" + f"[CONDITION: {condition}] - Processing skipped: Timeseries will be "
                              "empty when filtered to only include volumes from this specific condition. Possibly due to "
                              "TRs corresponding to the condition being removed by `dummy_scans` or filtered due to "
                              "exceeding threshold for `fd_threshold`.")
                continue

        if flagged is True:
            percentage_message = f"Percentage of volumes exceeding the threshold limit is {percentage*100}%"
            if condition: percentage_message = f"{percentage_message} for [CONDITION: {condition}]"
            warnings.warn(f"{subject_header}" + f"Processing skipped: Run flagged due to more than {outlier_limit*100}% "
                          f"of the volumes exceeding the framewise displacement (FD) threshold limit of {threshold}. "
                          f"{percentage_message}.")
            continue

        timeseries = _continue_extraction(nifti_file=nifti_file, mask_file=mask_file, confound_df=confound_df,
                                          confound_metadata_file=confound_metadata_file, condition=condition,
                                          signal_clean_info=signal_clean_info, dummy_scans=dummy_scans,tr=tr,
                                          parcel_approach=parcel_approach, flush_print=flush_print, verbose=verbose,
                                          censor_volumes=censor_volumes, scan_list=scan_list,
                                          subject_header=subject_header)

        if timeseries.shape[0] == 0:
            warnings.warn(f"{subject_header}" + f"Timeseries is empty for {run} and will not be appended to the "
                          "`subject_timeseries` dictionary.")
        else:
            subject_timeseries[subj_id].update({run_id: timeseries})

    if not subject_timeseries[subj_id]:
        warnings.warn(f"{sub_message.split(' | RUN:')[0]}]\n"
                      f"{'-'*len(sub_message.split(' | RUN:')[0])}\n"
                      "Processing skipped: No runs were extracted.")
        subject_timeseries = None

    return subject_timeseries

def _continue_extraction(nifti_file, mask_file, confound_df, confound_metadata_file, signal_clean_info,
                         dummy_scans, condition, tr, parcel_approach, flush_print, verbose, censor_volumes, scan_list,
                         subject_header):
    # Extract confound information of interest and ensure confound file does not contain NAs
    if signal_clean_info["use_confounds"]:
        # Extract first "n" numbers of specified WM and CSF components
        # acompcor extraction from fmriprep json files implementation from Michael Riedel & JulioAPeraza
        # at https://github.com/NBCLab/abcd_fmriprep-analysis/blob/5ffd3ba2d6e6d56344318abdd90c8f2f19a93101/analysis/rest/denoising.py#L111
        confound_names = copy.deepcopy(signal_clean_info["confound_names"])
        if confound_metadata_file:
            with open(confound_metadata_file, "r") as confounds_json:
                confound_metadata = json.load(confounds_json)

            acompcors = sorted([acompcor for acompcor in confound_metadata if "a_comp_cor" in acompcor])

            acompcors_CSF = [acompcor_CSF for acompcor_CSF in acompcors if
                                confound_metadata[acompcor_CSF]["Mask"] == "CSF"][0:signal_clean_info["n_acompcor_separate"]]
            acompcor_WM = [acompcor_WM for acompcor_WM in acompcors if
                            confound_metadata[acompcor_WM]["Mask"] == "WM"][0:signal_clean_info["n_acompcor_separate"]]

            confound_names.extend(acompcors_CSF + acompcor_WM)

        valid_confounds = []
        invalid_confounds = []
        for confound_name in confound_names:
            if "*" in confound_name:
                prefix = confound_name.split("*")[0]
                confounds_list = [col for col in confound_df.columns if col.startswith(prefix)]
            else:
                confounds_list = [col for col in confound_df.columns if col == confound_name]

            if confounds_list: valid_confounds.extend(confounds_list)
            else: invalid_confounds.extend([confound_name])

        if invalid_confounds:
            if verbose:
                print(f"{subject_header}" + f"The following confounds were not found: {', '.join(invalid_confounds)}.",
                      flush=flush_print)

        confounds = confound_df[valid_confounds]
        confounds = confounds.fillna(0)
        if verbose:
            print(f"{subject_header}" + "The following confounds will be for nuisance regression: "
                  f"{', '.join(list(confounds.columns))}.",flush=flush_print)

    # Create the masker for extracting time series
    masker = NiftiLabelsMasker(
        mask_img=mask_file,
        labels_img=parcel_approach[list(parcel_approach)[0]]["maps"],
        resampling_target='data',
        standardize=signal_clean_info["standardize"],
        detrend=signal_clean_info["detrend"],
        low_pass=signal_clean_info["low_pass"],
        high_pass=signal_clean_info["high_pass"],
        t_r=tr,
        smoothing_fwhm=signal_clean_info["fwhm"]
    )

    # Load and discard volumes if needed
    nifti_img = load_img(nifti_file)
    if dummy_scans:
        nifti_img = index_img(nifti_img, slice(dummy_scans, None))
        if signal_clean_info["use_confounds"]:
            confounds.drop(list(range(0,dummy_scans)),axis=0,inplace=True)

    # Extract timeseries
    if signal_clean_info["use_confounds"]: timeseries = masker.fit_transform(nifti_img, confounds=confounds)
    else: timeseries = masker.fit_transform(nifti_img)

    if censor_volumes and condition is None:
        timeseries = np.delete(timeseries, censor_volumes, axis=0)

    if condition:
        # If any out of bound indices, remove them instead of getting indexing error.
        if max(scan_list) > timeseries.shape[0] - 1:
            scan_list = _check_indices(scan_list,timeseries.shape[0],condition,subject_header)
        # Extract condition
        timeseries = timeseries[scan_list,:]

    return timeseries

def _check_indices(scan_list, arr_shape, condition, subject_header):
    warnings.warn(f"{subject_header}" + f"Max scan index for [CONDITION: {condition}] exceeds timeseries max index. "
                  f"Max condition index is {max(scan_list)}, while timeseries max index is {arr_shape - 1}. Timing may "
                  "be misaligned or repetition time incorrect. If intentional, ignore warning. Extracting indices for "
                  "condition only within timeseries range.")
    return [scan for scan in scan_list if scan in range(0,arr_shape)]



import copy, os
from typing import Union, Optional
import numpy as np
from .._utils import _convert_pickle_to_dict, _dicts_to_pickles

def merge_dicts(subject_timeseries_list: Union[list[dict[str, dict[str, np.ndarray]]], list[os.PathLike]],
                return_merged_dict: bool=True, return_reduced_dicts: bool=False,
                output_dir: Optional[Union[str, os.PathLike]]=None,
                file_names: Optional[list[str]]=None,
                save_reduced_dicts: bool=False) -> dict[str, dict[str, dict[str, np.ndarray]]]:
    """
    **Merge Subject Timeseries**

    Merge subject timeseries dictionaries or pickle files into the first dictionary or pickle file in the list.
    Repetition times/frames from the same subject and run are merged together. The merged dictionary will only
    include subjects that are present in all dictionaries.

    Parameters
    ----------
    subject_timeseries_list : :obj:`list[dict[str, dict[str, np.ndarray]]]` or :obj:`list[os.PathLike]`
        A list of dictionaries or pickle files containing the nested subject timeseries dictionary saved by the
        ``TimeSeriesExtractor`` class or a list of nested subject timeseries dictionaries produced by the
        ``TimeSeriesExtractor`` class. The first level of the nested dictionary must consist of the subject ID as a
        string, the second level must consist of the run numbers in the form of "run-#"
        (where # is the corresponding number of the run), and the last level must consist of the timeseries
        (as a numpy array) associated with that run. The structure is as follows:
        ::

            subject_timeseries = {
                    "101": {
                        "run-0": np.array([...]), # 2D array
                        "run-1": np.array([...]), # 2D array
                        "run-2": np.array([...]), # 2D array
                    },
                    "102": {
                        "run-0": np.array([...]), # 2D array
                        "run-1": np.array([...]), # 2D array
                    }
                }

    return_merged_dict : :obj:`bool`, default=True
        If True, returns the merged dictionary.

        ..versionchanged 0.15.0 from ``return_combined_dict`` to ``return_merged_dict``

    return_reduced_dicts : :obj:`bool`, default=False
        If True, returns the list of dictionaries provided with only the subjects present in the merged
        dictionary. The dictionaries are returned in the same order as listed in the ``subject_timeseries_list``
        parameter. The keys will be names "dict_#", with "#" indicating the index of the dictionary or pickle file
        in the ``subject_timeseries_list`` parameter.

    output_dir : :obj:`os.PathLike` or :obj:`None`, default=None
        Directory to save the merged or reduced dictionaries, as pickle files, to. The directory will be created
        if it does not exist. For the reduced dictionaries to be saved, ``save_reduced_dicts`` must be set to True.
        If ``save_reduced_dicts`` is False and ``output_dir`` is provided, only the merged dictionary will be saved.

    file_names : :obj:`list[str]` or :obj:`None`, default=None
        A list of names to save the dictionaries as if ``output_dir`` is provided. If ``save_reduced_dicts`` is False,
        only a list with a single name should be supplied, which will be used to save the merged dictionary. If
        ``save_reduced_dicts`` is True, then the length of the file_name must match the length of
        ``subject_timeseries_list`` plus an additional name for the merged dictionary. For instance, if for
        dictionaries or pickle files are provided in ``subject_timeseries_list`` and ``save_reduced_dicts`` is True,
        five names need to be provided. Additionally, the assignment of file names to dictionaries depends on the
        index position (file name in 0th index in the list will be assigned to the reduced version of the dictionary in
        the 0th index of ``subject_timeseries_list``. The last file name is always assigned to the merged dictionary.
        For this parameter, ``os.path.basename`` is used to get the basename of the files (if a full path is supplied)
        and ``os.path.splitext`` is used to ignore extensions. Default names are provided if this variable is None.

        ..versionchanged 0.15.0 from ``file_name`` to ``file_names``

    save_reduced_dicts : :obj:`bool` or None, default=False
        If True and the ``output_dir`` is provided, then the reduced dictionaries are saved.

        ..versionadded 0.15.0

    Returns
    -------
    `dict[str, dict[str, dict[str, np.ndarray]]]`
    """

    assert isinstance(subject_timeseries_list, list), "`subject_timeseries_list` must be a list."
    assert len(subject_timeseries_list) > 1, "Merging cannot be done with less than two dictionaries or files."

    if isinstance(subject_timeseries_list[0],dict): subject_timeseries_merged = subject_timeseries_list[0]
    else: subject_timeseries_merged = _convert_pickle_to_dict(pickle_file=subject_timeseries_list[0])

    # Get common subject ids
    subject_set = {}

    for curr_dict in subject_timeseries_list:
        if isinstance(curr_dict, str) and curr_dict.endswith(".pkl"):
            curr_dict = _convert_pickle_to_dict(pickle_file=curr_dict)
        if not subject_set: subject_set = set(curr_dict)
        subject_set = subject_set.intersection(list(curr_dict))

    # Order subjects
    intersect_subjects = sorted(list(subject_set))

    subject_timeseries_merged = {}

    for curr_dict in subject_timeseries_list:
        if isinstance(curr_dict, str) and curr_dict.endswith(".pkl"):
            curr_dict = _convert_pickle_to_dict(pickle_file=curr_dict)
        for subj_id in intersect_subjects:
            if subj_id not in subject_timeseries_merged: subject_timeseries_merged.update({subj_id: {}})
            # Get run names in the current iteration
            for curr_run in curr_dict[subj_id]:
                # If run is in merged dict, stack. If not, add
                if curr_run in subject_timeseries_merged[subj_id]:
                    subject_timeseries_merged[subj_id][curr_run] = np.vstack([subject_timeseries_merged[subj_id][curr_run],
                                                                                curr_dict[subj_id][curr_run]])
                else:
                    subject_timeseries_merged[subj_id].update({curr_run: curr_dict[subj_id][curr_run]})
            # Sort runs lexicographically, keys may be disordered if the first curr_dict does not contain the earliest run_id
            if list(subject_timeseries_merged[subj_id]) != sorted(subject_timeseries_merged[subj_id].keys()):
                subject_timeseries_merged[subj_id] = {run_id: subject_timeseries_merged[subj_id][run_id] for run_id
                                                        in sorted(subject_timeseries_merged[subj_id].keys())}

    modified_dicts = {}

    if return_reduced_dicts or (save_reduced_dicts and output_dir):
        for indx, curr_dict in enumerate(subject_timeseries_list):
            if "pkl" in curr_dict: curr_dict = _convert_pickle_to_dict(pickle_file=curr_dict)
            else: curr_dict = copy.deepcopy(curr_dict)
            if any([elem in subject_timeseries_merged for elem in curr_dict]):
                modified_dicts[f"dict_{indx}"] = {}
                for subj_id in subject_timeseries_merged:
                    if subj_id in curr_dict: modified_dicts[f"dict_{indx}"].update({subj_id : curr_dict[subj_id]})

    if return_merged_dict or output_dir: modified_dicts["merged"] = subject_timeseries_merged



    if output_dir:
        message = ("Length of `file_names` must be equal to 1 if `save_reduced_dicts`is False or the length of "
                   "`subject_timeseries_list` + 1 if `save_reduced_dicts` is True.")

        _dicts_to_pickles(output_dir=output_dir, dict_list=modified_dicts, call="merge", file_names=file_names,
                          message=message,save_reduced_dicts=save_reduced_dicts)

    if return_merged_dict or return_reduced_dicts: return modified_dicts



from ._capgetter import _CAPGetter
from ._cap2statmap import _cap2statmap
from ._heatmap_utils import _create_display, _save_contents
from ._create_labels import _create_node_labels
from ._run_kmeans import _run_kmeans



"""# A class which is responsible for accessing all TimeseriesExtractorGetter and to keep track of all attributes in TimeSeriesExtractor"""
import copy
import numpy as np
from .._check_parcel_approach import _check_parcel_approach
from .._pickle_to_dict import _convert_pickle_to_dict

class _TimeseriesExtractorGetter:
    def __init__(self):
        pass

    #### Exists upon initialization of TimeseriesExtractor
    @property
    def space(self):
        return self._space

    @property
    def signal_clean_info(self):
        return self._signal_clean_info

    @property
    def parcel_approach(self):
        return self._parcel_approach

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict):
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    ### Does not exists upon initialization of Timeseries Extractor

    # Exist when TimeSeriesExtractor.get_bold() used
    @property
    def task_info(self):
        return self._task_info if hasattr(self, "_task_info") else None

    # Gets initialized and populated in TimeSeriesExtractor.get_bold(),
    @property
    def subject_ids(self):
        return self._subject_ids if hasattr(self, "_subject_ids") else None

    @property
    def n_cores(self):
        return self._n_cores if hasattr(self, "_n_cores") else None

    # Gets initialized in TimeSeriesExtractor.get_bold(), gets populated when
    # TimeseriesExtractor._timeseries_aggregator gets called in TimeseriesExtractor._extract_timeseries
    @property
    def subject_timeseries(self):
        return self._subject_timeseries if hasattr(self, "_subject_timeseries") else None

    @subject_timeseries.setter
    def subject_timeseries(self, subject_dict):
        error_message = ("Must be a valid pickle file/subject timeseries should contain a nested dictionary where the "
                         "first level is the subject id, second level is the run number in the form of 'run-#', and the "
                         "final level is the timeseries as a numpy array.")
        if isinstance(subject_dict, str) and subject_dict.endswith(".pkl"):
            self._subject_timeseries = _convert_pickle_to_dict(subject_dict)
        elif isinstance(subject_dict, dict):
            first_level_indx = list(subject_dict)[0]
            if isinstance(subject_dict[first_level_indx], dict) and subject_dict[first_level_indx] and "run" in list(subject_dict[first_level_indx])[0]:
                run = list(subject_dict[first_level_indx])[0]
                if isinstance(subject_dict[first_level_indx][run],np.ndarray):
                    self._subject_timeseries = copy.deepcopy(subject_dict)
                else:
                    raise TypeError(error_message)
            else:
                raise TypeError(error_message)
        else:
            raise TypeError(error_message)



"""A class which is responsible for accessing all CAP metadata and to keep track of all attributes in CAP"""
import copy
from .._check_parcel_approach import _check_parcel_approach

class _CAPGetter:
    def __init__(self):
        pass

    ### Attributes exist when CAP initialized
    @property
    def groups(self):
        return self._groups

    @property
    def parcel_approach(self):
        return self._parcel_approach

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict):
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    ### Attributes exist when CAP.get_caps() used
    @property
    def n_clusters(self):
        return self._n_clusters if hasattr(self, "_n_clusters") else None

    @property
    def cluster_selection_method(self):
        return self._cluster_selection_method if hasattr(self, "_cluster_selection_method") else None

    @property
    def n_cores(self):
        return self._n_cores if hasattr(self, "_n_cores") else None

    @property
    def runs(self):
        return self._runs if hasattr(self, "_runs") else None

    @property
    def caps(self):
        return self._caps if hasattr(self, "_caps") else None

    @property
    def kmeans(self):
        return self._kmeans if hasattr(self, "_kmeans") else None

    @property
    def davies_bouldin(self):
        return self._davies_bouldin if hasattr(self, "_davies_bouldin") else None

    @property
    def silhouette_scores(self):
        return self._silhouette_scores if hasattr(self, "_silhouette_scores") else None

    @property
    def inertia(self):
        return self._inertia if hasattr(self, "_inertia") else None

    @property
    def variance_ratio(self):
        return self._variance_ratio if hasattr(self, "_variance_ratio") else None

    @property
    def optimal_n_clusters(self):
        return self._optimal_n_clusters if hasattr(self, "_optimal_n_clusters")else None

    @property
    def standardize(self):
        return self._standardize if hasattr(self, "_standardize") else None

    @property
    def means(self):
        return self._mean_vec if hasattr(self, "_mean_vec") else None

    @property
    def stdev(self):
        return self._stdev_vec if hasattr(self, "_stdev_vec") else None

    @property
    def concatenated_timeseries(self):
        return self._concatenated_timeseries if hasattr(self, "_concatenated_timeseries") else None

    # Generated in `caps2plot`
    @property
    def region_caps(self):
        return self._region_caps if hasattr(self, "_region_caps") else None

    @property
    def outer_products(self):
        return self._outer_products if hasattr(self, "_outer_product") else None

    @property
    def subject_table(self):
        return self._subject_table if hasattr(self, "_subject_table") else None

    @subject_table.setter
    def subject_table(self, subject_dict):
        if isinstance(subject_dict, dict):
            self._subject_table = copy.deepcopy(subject_dict)

    @property
    def cosine_similarity(self):
        return self._cosine_similarity if hasattr(self, "_cosine_similarity") else None



"""Internal function for checking the validity of parcel_approach."""
import copy, os, re, warnings
from nilearn import datasets
from ._pickle_to_dict import _convert_pickle_to_dict

def _check_parcel_approach(parcel_approach, call = "TimeseriesExtractor"):
    if isinstance(parcel_approach, str) and parcel_approach.endswith(".pkl"):
        parcel_approach = _convert_pickle_to_dict(parcel_approach)
    else:
        parcel_approach = copy.deepcopy(parcel_approach)
    valid_parcel_dict = {"Schaefer": {"n_rois" : 400, "yeo_networks": 7, "resolution_mm": 1},
                         "AAL": {"version": "SPM12"},
                         "Custom": {"maps": "/location/to/parcellation.nii.gz",
                                    "nodes": ["LH_Vis1", "LH_Vis2", "LH_Hippocampus",
                                              "RH_Vis1", "RH_Vis2", "RH_Hippocampus"],
                                    "regions": {"Vis" : {"lh": [0,1],
                                                          "rh": [3,4]},
                                                 "Hippocampus": {"lh": [2],
                                                                 "rh": [5]}}}}

    if not isinstance(parcel_approach,dict) or not list(parcel_approach)[0] in list(valid_parcel_dict):
        error_message = ("Please include a valid `parcel_approach` in one of the following dictionary formats for "
                         f"'Schaefer', 'AAL', or 'Custom':\n{valid_parcel_dict}")
        if not isinstance(parcel_approach,dict): raise TypeError(error_message)
        else: raise KeyError(error_message)

    if len(parcel_approach) > 1:
        raise ValueError("Only one parcellation approach can be selected. Example format of `parcel_approach`:\n"
                         f"{valid_parcel_dict}")

    if "Schaefer" in parcel_approach:
        if "n_rois" not in parcel_approach["Schaefer"]:
            warnings.warn("'n_rois' not specified in `parcel_approach`. Defaulting to 400 ROIs.")
            parcel_approach["Schaefer"].update({"n_rois": 400})

        if "yeo_networks" not in parcel_approach["Schaefer"]:
            warnings.warn("'yeo_networks' not specified in `parcel_approach`. Defaulting to 7 networks.")
            parcel_approach["Schaefer"].update({"yeo_networks": 7})

        if "resolution_mm" not in parcel_approach["Schaefer"]:
            warnings.warn("'resolution_mm' not specified in `parcel_approach`. Defaulting to 1mm.")
            parcel_approach["Schaefer"].update({"resolution_mm": 1})

        # Get atlas
        fetched_schaefer = datasets.fetch_atlas_schaefer_2018(n_rois=parcel_approach["Schaefer"]["n_rois"],
                                                              yeo_networks=parcel_approach["Schaefer"]["yeo_networks"],
                                                              resolution_mm=parcel_approach["Schaefer"]["resolution_mm"])
        parcel_approach["Schaefer"].update({"maps": fetched_schaefer.maps})
        network_name = "7Networks_" if parcel_approach["Schaefer"]["yeo_networks"] == 7 else "17Networks_"
        parcel_approach["Schaefer"].update({"nodes": [label.decode().split(network_name)[-1]
                                                      for label in fetched_schaefer.labels]})
        # Get node networks
        parcel_approach["Schaefer"].update({"regions": list(dict.fromkeys([re.split("LH_|RH_", node)[-1].split("_")[0]
                                                                           for node in parcel_approach["Schaefer"]["nodes"]]))})

    elif "AAL" in parcel_approach:
        if "version" not in parcel_approach["AAL"]:
            warnings.warn("'version' not specified in `parcel_approach`. Defaulting to 'SPM12'.")
            parcel_approach["AAL"].update({"version": "SPM12"})

        # Get atlas
        fetched_aal = datasets.fetch_atlas_aal(version=parcel_approach["AAL"]["version"])
        parcel_approach["AAL"].update({"maps": fetched_aal.maps})
        parcel_approach["AAL"].update({"nodes": [label for label in fetched_aal.labels]})
        # Get node networks
        parcel_approach["AAL"].update({"regions": list(dict.fromkeys([node.split("_")[0]
                                                                      for node in parcel_approach["AAL"]["nodes"]]))})

    else:
        custom_example = {"Custom": valid_parcel_dict["Custom"]}
        if call  == "TimeseriesExtractor" and "maps" not in parcel_approach["Custom"]:
            raise ValueError("For `Custom` parcel_approach, a nested key-value pair containing the key 'maps' with the "
                             "value being a string specifying the location of the parcellation is needed. Example:\n"
                             f"{custom_example}")
        check_subkeys = ["nodes" in parcel_approach["Custom"], "regions" in parcel_approach["Custom"]]
        if not all(check_subkeys):
            missing_subkeys = [["nodes", "regions"][x] for x,y in enumerate(check_subkeys) if y is False]
            error_message = f"The following sub-keys haven't been detected {missing_subkeys}"
            if call == "TimeseriesExtractor":
                warnings.warn(f"{error_message}. These labels are not needed for timeseries extraction but are needed "
                              "for future timeseries or CAPs plotting.")
            else:
                raise ValueError(f"{error_message}. Certain sub-keys are needed for plotting. Please check the function "
                                 "docstring for the required sub-keys and reassign `parcel_approach` using "
                                 f"`self.parcel_approach`. Please refer to the example structure:\n{custom_example}")
        if call == "TimeseriesExtractor" and not os.path.isfile(parcel_approach["Custom"]["maps"]):
            raise FileNotFoundError("Please specify the location to the custom parcellation to be used.")

    return parcel_approach
